from typing import Any, List
from h1st.model.ml_model import MLModel

from h1st.model.predictive_model import PredictiveModel

class Ensemble(PredictiveModel):
    """
    Base ensemble class.
    """
