from h1st.model.predictive_model import PredictiveModel


class Teacher(PredictiveModel):
    """
    Teacher Model in Oracle framework
    """
    pass


class TeacherModeler():
    """
    Knowledge Generalization Modeler
    """
    pass
