from typing import List
from h1st.model.predictive_model import PredictiveModel
from h1st.model.modeler import Modeler
from h1st.model.ml_modeler import MLModeler

class EnsembleModeler(Modeler):
    pass
